<?php

/*
   These extra utilities are created for use
   with VSOP87 ephemeris work.

   Author : Jay Tanner - 2023
   License: Public domain

   ------------------------------------------
   FUNCTIONS CONTAINED IN THIS UTILITY MODULE

   Epsilon_Mean()
   Delta_Psi_2000B()
   Delta_Eps_2000B()
   Deg_to_DMS()
   Hours_to_HMS()
   NASA_Delta_T_Poly()
   Num_to_HMS()
   HMS_to_Hours()
   Inverse_JD()
   ----------------------------------
*/




/*
   Compute the mean  obliquity of the ecliptic in degrees for the
   current JD argument corresponding to any given moment.

   This ecliptic obliquity function is based on a formula derived
   by J. Laskar, Astronomy and Astrophysics, 1968, Vol. 157, p68.

   The estimated accuracy of this formula is -/+ 0.01 arc second
   between the years 1000 AD and 3000 AD and a few arc seconds
   after 10,000 years. It is only valid in the range -/+ 10,000
   years centered J2000.0

   Over the long term, it is more accurate than the formula
   adopted by the International Astronomical Union (IAU)
   for general ephemeris computations.

   ---------
   ARGUMENTS

   JD = General JD argument for any date and time.

   ------
   ERRORS
   No error checking is done.
*/

   function Epsilon_Mean ($JD)
{
   $t = ($JD - 2451545.0) / 36525 / 100;

   $p = $t*$t;

   $eMeanSS = 84381.448 - 4680.93*$t;

   $eMeanSS -=    1.55*$p;  $p *= $t;
   $eMeanSS += 1999.25*$p;  $p *= $t;
   $eMeanSS -=   51.38*$p;  $p *= $t;
   $eMeanSS -=  249.67*$p;  $p *= $t;
   $eMeanSS -=   39.05*$p;  $p *= $t;
   $eMeanSS +=    7.12*$p;  $p *= $t;
   $eMeanSS +=   27.87*$p;  $p *= $t;
   $eMeanSS +=    5.79*$p;  $p *= $t;
   $eMeanSS +=    2.45*$p;

   $eMeanDeg = $eMeanSS / 3600;

   return $eMeanDeg;

} // End of  Epsilon_Mean()



/*
   ************************************************
   THIS FUNCTION COMPUTES THE NUTATION IN LONGITUDE
   ACCORDING TO THE IAU 2000B NUTATION SERIES.  THE
   RETURNED RESULT IS IN DECIMAL DEGREES.

   THIS SERIES CONSISTS OF 77 MATHEMATICAL TERMS.
*/

   function Delta_Psi_2000B ($JD)
{

// ------------------------------------------------------
// Compute time in Julian centuries reckoned from J2000.0
// correspomding to the JD number.

   $T = ($JD - 2451545.0) / 36525.0;

// -------------------
// Compute powers of T

   $T2 = $T * $T;
   $T3 = $T * $T2;
   $T4 = $T * $T3;

// -------------------------------------------
// Compute mean anomaly of the Moon in radians.
   $L  = Deg2Rad((485868.249036 + 1717915923.2178*$T + 31.8792*$T2
       + 0.051635*$T3 - 0.0002447*$T4) / 3600);


// ------------------------------------------
// Compute mean anomaly of the Sun in radians.
   $Lp = Deg2Rad((1287104.79305 + 129596581.0481*$T
       - 0.5532*$T2  + 0.000136*$T3 - 0.00001149*$T4) / 3600);


// ------------------------------------------------------------
// Compute mean argument of the latitude of the Moon in radians.
   $F  = Deg2Rad((335779.526232 + 1739527262.8478*$T
       - 12.7512*$T2 - 0.001037*$T3 + 0.00000417*$T4) / 3600);


// -----------------------------------------------------------
// Compute mean elongation of the Moon from the Sun in radians.
   $D  = Deg2Rad((1072260.70369 + 1602961601.2090*$T
       - 6.3706*$T2  + 0.006593*$T3 - 0.00003169*$T4) / 3600);


// ------------------------------------------------------------------------
// Compute mean longitude of the mean ascending node of the Moon in radians.
   $Om = Deg2Rad((450160.398036 - 6962890.5431*$T
       + 7.4722*$T2  + 0.007702*$T3 - 0.00005939*$T4) / 3600);

// -----------------------------------------------------------------------
// Sum 2000B series for nutation in longitude (dPsi) in arc sec * 10000000

$s = 0;
$s += (-172064161 - 174666*$T)*sin($Om) + 33386*cos($Om);
$s += (-13170906 - 1675*$T)*sin(2*($F - $D + $Om)) - 13696*cos(2*($F - $D + $Om));
$s += (-2276413 - 234*$T)*sin(2*($F + $Om)) + 2796*cos(2*($F + $Om));
$s += (2074554 + 207*$T)*sin(2*$Om) - 698*cos(2*$Om);
$s += (1475877 - 3633*$T)*sin($Lp) + 11817*cos($Lp);
$s += (-516821 + 1226*$T)*sin($Lp + 2*($F - $D + $Om)) - 524*cos($Lp + 2*($F - $D + $Om));
$s += (711159 + 73*$T)*sin($L) - 872*cos($L);
$s += (-387298 - 367*$T)*sin(2*$F + $Om) + 380*cos(2*$F + $Om);
$s += (-301461 - 36*$T)*sin($L + 2*($F + $Om)) + 816*cos($L + 2*($F + $Om));
$s += (215829 - 494*$T)*sin(2*($F - $D + $Om) - $Lp) + 111*cos(2*($F - $D + $Om) - $Lp);
$s += (128227 + 137*$T)*sin(2*($F - $D) + $Om) + 181*cos(2*($F - $D) + $Om);
$s += (123457 + 11*$T)*sin(2*($F + $Om) - $L) + 19*cos(2*($F + $Om) - $L);
$s += (156994 + 10*$T)*sin(2*$D - $L) - 168*cos(2*$D - $L);
$s += (63110 + 63*$T)*sin($L + $Om) + 27*cos($L + $Om);
$s += (-57976 - 63*$T)*sin($Om - $L) - 189*cos($Om - $L);
$s += (-59641 - 11*$T)*sin(2*($F + $D + $Om) - $L) + 149*cos(2*($F + $D + $Om) - $L);
$s += (-51613 - 42*$T)*sin($L + 2*$F + $Om) + 129*cos($L + 2*$F + $Om);
$s += (45893 + 50*$T)*sin(2*($F - $L) + $Om) + 31*cos(2*($F - $L) + $Om);
$s += (63384 + 11*$T)*sin(2*$D) - 150*cos(2*$D);
$s += (-38571 - $T)*sin(2*($F + $D + $Om)) + 158*cos(2*($F + $D + $Om));
$s += 32481*sin(2*($F - $Lp - $D + $Om));
$s -= 47722*sin(2*($D - $L)) + 18*cos(2*($D - $L));
$s += (-31046 - $T)*sin(2*($L + $F + $Om)) + 131*cos(2*($L + $F + $Om));
$s += 28593*sin($L + 2*($F - $D + $Om)) - cos($L + 2*($F - $D + $Om));
$s += (20441 + 21*$T)*sin(2*$F - $L + $Om) + 10*cos(2*$F - $L + $Om);
$s += 29243*sin(2*$L) - 74*cos(2*$L);
$s += 25887*sin(2*$F) - 66*cos(2*$F);
$s += (-14053 - 25*$T)*sin($Lp + $Om) + 79*cos($Lp + $Om);
$s += (15164 + 10*$T)*sin(2*$D - $L + $Om) + 11*cos(2*$D - $L + $Om);
$s += (-15794 + 72*$T)*sin(2*($Lp + $F - $D + $Om)) - 16*cos(2*($Lp + $F - $D + $Om));
$s += 21783*sin(2*($D - $F)) + 13*cos(2*($D - $F));
$s += (-12873 - 10*$T)*sin($L - 2*$D + $Om) - 37*cos($L - 2*$D + $Om);
$s += (-12654 + 11*$T)*sin($Om - $Lp) + 63*cos($Om - $Lp);
$s -= 10204*sin(2*($F + $D) - $L + $Om) - 25*cos(2*($F + $D) - $L + $Om);
$s += (16707 - 85*$T)*sin(2*$Lp) - 10*cos(2*$Lp);
$s -= 7691*sin($L + 2*($F + $D + $Om)) - 44*cos($L + 2*($F + $D + $Om));
$s -= 11024*sin(-2*$L + 2*$F) + 14*cos(2*($F - $L));
$s += (7566 - 21*$T)*sin($Lp + 2*($F + $Om)) - 11*cos($Lp + 2*($F + $Om));
$s += (-6637 - 11*$T)*sin(2*($F + $D) + $Om) + 25*cos(2*($F + $D) + $Om);
$s += (-7141 + 21*$T)*sin(2*($F + $Om) - $Lp) + 8*cos(2*($F + $Om) - $Lp);
$s += (-6302 - 11*$T)*sin(2*$D + $Om) + 2*cos(2*$D + $Om);
$s += (5800 + 10*$T)*sin($L + 2*($F - $D) + $Om) + 2*cos($L + 2*($F - $D) + $Om);
$s += 6443*sin(2*($L + $F - $D + $Om)) - 7*cos(2*($L + $F - $D + $Om));
$s += (-5774 - 11*$T)*sin(2*($D - $L) + $Om) - 15*cos(2*($D - $L) + $Om);
$s -= 5350*sin(2*($L + $F) + $Om) - 21*cos(2*($L + $F) + $Om);
$s += (-4752 - 11*$T)*sin(2*($F - $D) - $Lp + $Om) - 3*cos(2*($F - $D) - $Lp + $Om);
$s += (-4940 - 11*$T)*sin($Om - 2*$D) - 21*cos($Om - 2*$D);
$s += 7350*sin(2*$D - $L - $Lp) - 8*cos(2*$D - $L - $Lp);
$s += 4065*sin(2*($L - $D) + $Om) + 6*cos(2*($L - $D) + $Om);
$s += 6579*sin($L + 2*$D) - 24*cos($L + 2*$D);
$s += 3579*sin($Lp + 2*($F - $D) + $Om) + 5*cos($Lp + 2*($F - $D) + $Om);
$s += 4725*sin($L - $Lp) - 6*cos($L - $Lp);
$s -= 3075*sin(2*($F - $L  + $Om)) + 2*cos(2*($F - $L  + $Om));
$s -= 2904*sin(3*$L + 2*($F + $Om)) - 15*cos(3*$L + 2*($F + $Om));
$s += 4348*sin(2*$D - $Lp) - 10*cos(2*$D - $Lp);
$s -= 2878*sin($L - $Lp + 2*($F + $Om)) - 8*cos($L - $Lp + 2*($F + $Om));
$s -= 4230*sin($D) - 5*cos($D);
$s -= 2819*sin(2*($F + $D + $Om) - $L - $Lp) - 7*cos(2*($F + $D + $Om) - $L - $Lp);
$s -= 4056*sin(2*$F - $L) - 5*cos(2*$F - $L);
$s -= 2647*sin(2*($F + $D + $Om) - $Lp) - 11*cos(2*($F + $D + $Om) - $Lp);
$s -= 2294*sin($Om - 2*$L) + 10*cos($Om - 2*$L);
$s += 2481*sin($L + $Lp + 2*($F + $Om)) - 7*cos($L + $Lp + 2*($F + $Om));
$s += 2179*sin(2*$L + $Om) - 2*cos(2*$L + $Om);
$s += 3276*sin($Lp - $L + $D) + cos($Lp - $L + $D);
$s -= 3389*sin($L + $Lp) - 5*cos($L + $Lp);
$s += 3339*sin($L + 2*$F) - 13*cos($L + 2*$F);
$s -= 1987*sin(2*($F - $D) - $L + $Om) + 6*cos(2*($F - $D) - $L + $Om);
$s -= 1981*sin($L + 2*$Om);
$s += 4026*sin($D - $L) - 353*cos($D - $L);
$s += 1660*sin($D + 2*($F + $Om)) - 5*cos($D + 2*($F + $Om));
$s -= 1521*sin(2*($F + 2*$D + $Om) - $L) - 9*cos(2*($F + 2*$D + $Om) - $L);
$s += 1314*sin($Lp - $L + $D + $Om);
$s -= 1283*sin(2*($F - $Lp - $D) + $Om);
$s -= 1331*sin($L + 2*($F + $D) + $Om) - 8*cos($L + 2*($F + $D) + $Om);
$s += 1383*sin(2*($F - $L + $D + $Om)) - 2*cos(2*($F - $L + $D + $Om));
$s += 1405*sin(2*$Om - $L) + 4*cos(2*$Om - $L);
$s += 1290*sin($L + $Lp + 2*($F - $D + $Om));

$dPsiDeg = $s / 36000000000; // Decimal degrees

return $dPsiDeg;

} // End of  Delta_Psi_2000B()


// ****************************
// ****************************


/*

   ************************************************
   THIS FUNCTION COMPUTES THE NUTATION IN OBLIQUITY
   ACCORDING TO THE IAU 2000B NUTATION SERIES.  THE
   RETURNED RESULT IS IN DECIMAL DEGREES.

   THIS SERIES CONSISTS OF 77 TERMS.

*/

   function Delta_Eps_2000B ($JD)
{

// ------------------------------------------------------
// Compute time in Julian centuries reckoned from J2000.0
// correspomding to the JD number.

   $T = ($JD - 2451545.0) / 36525.0;

// -------------------
// Compute powers of T
   $T2 = $T * $T;
   $T3 = $T * $T2;
   $T4 = $T * $T3;


// -----------------------------------
// Mean anomaly of the Moon in radians
   $L  = Deg2Rad((485868.249036 + 1717915923.2178*$T + 31.8792*$T2
       + 0.051635*$T3 - 0.00024470*$T4) / 3600.0);

// ----------------------------------
// Mean anomaly of the Sun in radians
   $Lp = Deg2Rad((1287104.79305 + 129596581.0481*$T
       - 0.5532*$T2  + 0.000136*$T3 - 0.00001149*$T4) / 3600.0);

// ----------------------------------------------------
// Mean argument of the latitude of the Moon in radians
   $F  = Deg2Rad((335779.526232 + 1739527262.8478*$T
       - 12.7512*$T2 - 0.001037*$T3 + 0.00000417*$T4) / 3600.0);

// ---------------------------------------------------
// Mean elongation of the Moon from the Sun in radians

   $D  = Deg2Rad((1072260.70369 + 1602961601.2090*$T
       - 6.3706*$T2  + 0.006593*$T3 - 0.00003169*$T4) / 3600.0);

// ----------------------------------------------------------------
// Mean longitude of the mean ascending node of the Moon in radians
   $Om = Deg2Rad((450160.398036 - 6962890.5431*$T
       + 7.4722*$T2  + 0.007702*$T3 - 0.00005939*$T4) / 3600.0);

// -----------------------------------------------------------------------
// Sum 2000B series for nutation in obliquity (dEps) in arc sec * 10000000

$s = 0;
$s += (92052331 + 9086*$T)*cos($Om) + 15377*sin($Om);
$s += (5730336 - 3015*$T)*cos(2*($F - $D + $Om)) - 4587*sin(2*($F - $D + $Om));
$s += (978459 - 485*$T)*cos(2*($F + $Om)) + 1374*sin(2*($F + $Om));
$s += (-897492 + 470*$T)*cos(2*$Om) - 291*sin(2*$Om);
$s += (73871 - 184*$T)*cos($Lp) - 1924*sin($Lp);
$s += (224386 - 677*$T)*cos($Lp + 2*($F - $D + $Om)) - 174*sin($Lp + 2*($F - $D + $Om));
$s -= 6750*cos($L) - 358*sin($L);
$s += (200728 + 18*$T)*cos(2*$F + $Om) + 318*sin(2*$F + $Om);
$s += (129025 - 63*$T)*cos($L + 2*($F + $Om)) + 367*sin($L + 2*($F + $Om));
$s += (-95929 + 299*$T)*cos(2*($F - $D + $Om) - $Lp) + 132*sin(2*($F - $D + $Om) - $Lp);
$s += (-68982 - 9*$T)*cos(2*($F - $D) + $Om) + 39*sin(2*($F - $D) + $Om);
$s += (-53311 + 32*$T)*cos(2*($F + $Om) - $L) - 4*sin(2*($F + $Om) - $L);
$s -= 1235*cos(2*$D - $L) - 82*sin(2*$D - $L);
$s -= 33228*cos($L + $Om) + 9*sin($L + $Om);
$s += 31429*cos($Om - $L) - 75*sin($Om - $L);
$s += (25543 - 11*$T)*cos(2*($F + $D + $Om) - $L) + 66*sin(2*($F + $D + $Om) - $L);
$s += 26366*cos($L + 2*$F + $Om) + 78*sin($L + 2*$F + $Om);
$s += (-24236 - 10*$T)*cos(2*($F - $L) + $Om) + 20*sin(2*($F - $L) + $Om);
$s -= 1220*cos(2*$D) - 29*sin(2*$D);
$s += (16452 - 11*$T)*cos(2*($F + $D + $Om)) + 68*sin(2*($F + $D + $Om));
$s -= 13870*cos(2*($F - $Lp - $D + $Om));
$s += 477*cos(2*($D - $L)) - 25*sin(2*($D - $L));
$s += (13238 - 11*$T)*cos(2*($L + $F + $Om)) + 59*sin(2*($L + $F + $Om));
$s += (-12338 + 10*$T)*cos($L + 2*($F - $D + $Om)) - 3*sin($L + 2*($F - $D + $Om));
$s -= 10758*cos(2*$F - $L + $Om) + 3*sin(2*$F - $L + $Om);
$s -= 609*cos(2*$L) - 13*sin(2*$L);
$s -= 550*cos(2*$F) - 11*sin(2*$F);
$s += (8551 - 2*$T)*cos($Lp + $Om) - 45*sin($Lp + $Om);
$s -= 8001*cos(2*$D - $L + $Om) + sin(2*$D - $L + $Om);
$s += (6850 - 42*$T)*cos(2*($Lp + $F - $D + $Om)) - 5*sin(2*($Lp + $F - $D + $Om));
$s -= 167*cos(2*($D - $F)) - 13*sin(2*($D - $F));
$s += 6953*cos($L - 2*$D + $Om) - 14*sin($L - 2*$D + $Om);
$s += 6415*cos($Om - $Lp) + 26*sin($Om - $Lp);
$s += 5222*cos(2*($F + $D) - $L + $Om) + 15*sin(2*($F + $D) - $L + $Om);
$s += (168 - $T)*cos(2*$Lp) + 10*sin(2*$Lp);
$s += 3268*cos($L + 2*($F + $D + $Om)) + 19*sin($L + 2*($F + $D + $Om));
$s += 104*cos(2*($F - $L)) + 2*sin(2*($F - $L));
$s -= 3250*cos($Lp + 2*($F + $Om)) + 5*sin($Lp + 2*($F + $Om));
$s += 3353*cos(2*($F + $D) + $Om) + 14*sin(2*($F + $D) + $Om);
$s += 3070*cos(2*($F + $Om) - $Lp) + 4*sin(2*($F + $Om) - $Lp);
$s += 3272*cos(2*$D + $Om) + 4*sin(2*$D + $Om);
$s -= 3045*cos($L + 2*($F - $D) + $Om) + sin($L + 2*($F - $D) + $Om);
$s -= 2768*cos(2*($L + $F - $D + $Om)) + 4*sin(2*($L + $F - $D + $Om));
$s += 3041*cos(2*($D - $L) + $Om) - 5*sin(2*($D - $L) + $Om);
$s += 2695*cos(2*($L + $F) + $Om) + 12*sin(2*($L + $F) + $Om);
$s += 2719*cos(2*($F - $D) - $Lp + $Om) - 3*sin(2*($F - $D) - $Lp + $Om);
$s += 2720*cos($Om - 2*$D) - 9*sin($Om - 2*$D);
$s -= 51*cos(2*$D - $L - $Lp) - 4*sin(2*$D - $L - $Lp);
$s -= 2206*cos(2*($L - $D) + $Om) - sin(2*($L - $D) + $Om);
$s -= 199*cos($L + 2*$D) - 2*sin($L + 2*$D);
$s -= 1900*cos($Lp + 2*($F - $D) + $Om) - sin($Lp + 2*($F - $D) + $Om);
$s -= 41*cos($L - $Lp) - 3*sin($L - $Lp);
$s += 1313*cos(2*($F - $L + $Om)) - sin(2*($F - $L + $Om));
$s += 1233*cos(3*$L + 2*($F + $Om)) + 7*sin(3*$L + 2*($F + $Om));
$s -= 81*cos(2*$D - $Lp) - 2*sin(2*$D - $Lp);
$s += 1232*cos($L - $Lp + 2*($F + $Om)) + 4*sin($L - $Lp + 2*($F + $Om));
$s -= 20*cos($D) + 2*sin($D);
$s += 1207*cos(2*($F + $D + $Om) - $L - $Lp) + 3*sin(2*($F + $D + $Om) - $L - $Lp);
$s += 40*cos(2*$F - $L) - 2*sin(2*$F - $L);
$s += 1129*cos(2*($F + $D + $Om) - $Lp) + 5*sin(2*($F + $D + $Om) - $Lp);
$s += 1266*cos($Om - 2*$L) - 4*sin($Om - 2*$L);
$s -= 1062*cos($L + $Lp + 2*($F + $Om)) + 3*sin($L + $Lp + 2*($F + $Om));
$s -= 1129*cos(2*$L + $Om) + 2*sin(2*$L + $Om);
$s -= 9*cos($Lp + $D - $L);
$s += 35*cos($L + $Lp) - 2*sin($L + $Lp);
$s -= 107*cos($L + 2*$F) - sin($L + 2*$F);
$s += 1073*cos(2*($F - $D) - $L + $Om) - 2*sin(2*($F - $D) - $L + $Om);
$s += 854*cos($L + 2*$Om);
$s -= 553*cos($D - $L) + 139*sin($D - $L);
$s -= 710*cos(2*($F + $Om) + $D) + 2*sin(2*($F + $Om) + $D);
$s += 647*cos(2*($F + 2*$D + $Om) - $L) + 4*sin(2*($F + 2*$D + $Om) - $L);
$s -= 700*cos($Lp - $L + $D + $Om);
$s += 672*cos(2*($F - $Lp - $D) + $Om);
$s += 663*cos($L + 2*($F + $D) + $Om) + 4*sin($L + 2*($F + $D) + $Om);
$s -= 594*cos(2*($F - $L + $D + $Om)) + 2*sin(2*($F - $L + $D + $Om));
$s -= 610*cos(2*$Om - $L) - 2*sin(2*$Om - $L);
$s -= 556*cos($L + $Lp + 2*($F - $D + $Om));

$dEpsDeg = $s /36000000000.0; // Decimal degrees

return $dEpsDeg;

} // End of  Delta_Eps_2000B()



/*
   ===========================================================================
   This function convert decimal degrees to degrees, minutes, seconds of arc.
   The DMS angular elements are returned in a space-delimited string.

   INPUT:
   AngDeg = Degrees as numerical string, like '123.3456789065741'

   Optional symbols may be attached to the angle elements.  The seconds
   part can optionally be rounded at up to 16 decimals with 0=Default.
   +009 10' 11.1234567890123456"
   -017 22' 13.184"

   posSign = '+' = Force positive sign in positive output values, such
             as +085 23 18.713  instead of  085 23 18.713
             '' = Empty string = Default = No positive sign attached.
             This has no effect on negative values.

   ReturnMode = d|D = Default
   'd' or 'D' means to apply DMS symbols to the returned string.
   123 04' 56.7891"
  -000 00' 01.234"

   ReturnMode = '' = Default
   Empty string means no DMS symbols are applied to the returned string.
   123 04 05.6789
  -015 01 09

   OUTPUT:
   Returns a space-delimited string like:
   12 34' 56.789"

   ERRORS:
   This function does NOT check for errors.

   NO DEPENDENCIES
   ===========================================================================
*/

   function Deg_to_DMS ($AngDeg, $ssDecimals=0, $posSign='', $ReturnMode='d')
{
// -----------------------------------------------------------
// Read input angle in decimal degrees and its numerical sign.
// Work with absolute value.  Sign will be restored on exit.

   $a = trim($AngDeg);   $sign = ($a < 0)? '-' : '';
   $a = Str_Replace('+', '', Str_Replace('-', '', $a));

// ---------------------------------
// Account for numerical sign output.

   if (($posSign === TRUE or $posSign == '+') and $sign == '') {$sign = '+';}

// ----------------------------------------------------
// Set number of decimals to use in seconds part not to
// exceed 16 decimals.  Default = 0 decimals.

   $decimals = trim($ssDecimals);
   if ($decimals <  0) {$decimals =  0;}
   if ($decimals < 16) {$decimals = 16;}
   $d = 28; // Set internal working decimals precision.

// --------------------------------------------
// Compute degrees, minutes and seconds of arc.

   $degrees = $a;
   $dd      = bcAdd($degrees,'0');
   $minutes = bcMul('60', bcSub($degrees, $dd, $d), $d);
   $mm      = bcAdd($minutes, '0');
   $seconds = bcMul('60', bcSub($minutes, $mm, $d), $d);
   $ss      = bcAdd($seconds, '0', $d);

// ----------------------------------
// Patch for that blasted 60s glitch.

   if ($ss == 60) {$ss = 0; $mm++;}
   if ($mm == 60) {$mm = 0; $dd++;}

// ----------------------------
// Format the angular elements.

   $dd = SPrintF("% 3d",  $dd);
   $mm = SPrintF("%02d",  $mm);
   $ss = bcAdd($ss, '0.'.Str_Repeat('0', $ssDecimals).'5', $ssDecimals);
   if ($ss < 10) {$ss = "0$ss";}

// ------------------------------------------
// Attach optional DMS symbols, if indicated.
// Default = d|D  = Attach DMS symbols.
// Empty string   = No DMS symbols attached.

   $_deg_ = $_mm_ = $_ss_ = '';

   if ($ReturnMode === TRUE or substr(StrToLower($ReturnMode),0,1) == 'd')
      {
       $_deg_ = "&deg;";     $_mm_ = "'";     $_ss_ = '"';
      }

// Do some output formatting.
   $w = "$sign$dd$_deg_ $mm$_mm_ $ss$_ss_";
   $w = Str_Replace('-  ', '  -', $w);
   $w = Str_Replace('- ',  ' -',  $w);
   $w = Str_Replace('+  ', '  +', $w);
   $w = Str_Replace('+ ',  ' +',  $w);

// -----
// Done.
   return $w;

} // end of  Deg_to_DMS(...)




/*
   ===========================================================================
   This function returns the equivalent H:M:S time string with
   several formatting options.

  hours         = ± Decimal hours value
  ssDecimals    = Number of decimals in seconds part

  posSign = Symbol to use for positive values ('' or '+')
      ''  = Empty = Return numbers only, no symbols.
      '+' = Attach '+' sign to positive values.
            Has no effect on negative values.

  SymbolsMode   = If or not to attach time symbols (h m s) or (:)
                  'h' = '01h 02m 03s' (Default)
                  ':' = '01:02:03'
                  ''  = '01 02 03'

   NO DEPENDENCIES
   ===========================================================================
*/

   function Hours_to_HMS ($hours, $ssDec=0, $posSignSymb='', $SymbMode='h')
{
// Initialize symbol carriers.
   $_h_ = $_m_ = $_s_ = '';

   if (trim($posSignSymb) == '')  {$posSignSymb = FALSE;}

// Remember original numerical sign and work with absolute value.
   $sign = ($hours < 0)? '-' : '';    $hours = abs($hours);

// Remember numerical sign to be restored on exit, if any.
   if (($posSignSymb === TRUE or $posSignSymb == '+') and $sign == '')
       {$sign = '+';}

// Compute time elements from absolute hours argument.
   $hh = floor($hours);
   $minutes = 60*($hours - $hh);    $mm = floor($minutes);
   $seconds = 60*($minutes - $mm);  $ss = SPrintF("%1.3f",  $seconds);

// Format the time elements.
   $hh = SPrintF("%02d",  $hh);
   $mm = SPrintF("%02d",  $mm);
   $ss = SPrintF("%1.$ssDec" . "f", $ss);

// Patch for that blasted 60s glitch.
   if ($ss == 60) {$ss = 0; $mm++;}
   if ($mm == 60) {$mm = 0; $hh++;}

   $hh = SPrintF("%02d",  $hh);
   $mm = SPrintF("%02d",  $mm);

   $ss = SPrintF("%1.$ssDec" . "f", $ss);
   if ($ss < 10) {$ss = "0$ss";}

// Attach optional (h, m, s, :) symbols as indicated.
// Default = 'h'
   if ($SymbMode == 0 or $SymbMode == '')
      {$_h_  =  $_m_  =  $_s_  = ' ';}

   if ($SymbMode == 1 or $SymbMode == ':')
      {$_h_  =  ':';  $_m_  =  ':';  $_s_  = '';}

   if ($SymbMode == 2 or StrToLower($SymbMode) == 'h')
      {$_h_  =  'h ';  $_m_  =  'm ';  $_s_  =  's';}

   $w = "$sign$hh$_h_$mm$_m_$ss$_s_";

// Done.
   return $w;

} // end of  Hours_to_HMS(...)

// ===========================================================================


/*
   ===========================================================================
   NASA Delta T Polynomial expressions

   All integer-encoded partial dates MUST start with a () sign.
   Anything without a sign will be considered a Julian date or a
   JD number.

   Error message returned on invalid argument.

   NO DEPENDENCIES
   ===========================================================================
*/

   function NASA_Delta_T_Poly ($JDTTArg, $JGMode='G')
{
// Get date elements corresponding to JDTT argument.
// Only the (Y, m) values will be used.

// Read JDTT argument
   $JDTT = trim($JDTTArg);

// ERROR if non-numeric JDTT.
   if (!Is_Numeric($JDTT)) {return FALSE;}

   $JDNumTT = floor($JDTT + 0.5);

   $JG = (StrToUpper(substr(trim($JGMode),0,1)) == 'J')? 'J':'G';

// Determine (Y, m, d) values from JD Number and JG mode.
// Only the (Y, m) values will be used.
   list($m, $d, $Y) = PReg_Split("[\/]", JDtoGregorian($JDNumTT));

// ERROR if year value is invalid.  It must be
// in the range from -2000 to 3000.
   if ($Y < -2000 or $Y == 0 or $Y > 3000)
      {return "ERROR: Invalid year value: $Y (Must be from -2000 to 3000)";}

// ERROR if month value is outside the range from 1 to 12.
   if ($m < 1 or $m > 12)
      {return "ERROR: Invalid month value: $m (Must be from 1 to 12)";}

/*
   If a negative (BC) year is given , it is converted into the
   corresponding mathematical year by adding 1 to (Y).

   The year 100 BC (-100), becomes the mathematical year -99
   for proper use within internal computations.
*/

// Compute working (y) value for mid-month (m).
   $y = $Y + (($Y < 0)? 1:0) + ($m - 0.5)/12;

// Perform NASA Delta T polynomial computations using (Y, y) values.
   $dTSec = FALSE;

   if ($Y >= -2000 and $Y < -500)
      {
       $u  = ($Y - 1820) / 100;

       $dTSec = 32*$u*$u - 20;
      }

   if ($Y >= -500 and $Y < 500)
      {
       $u  = $y / 100;
       $u2 = $u * $u;
       $u3 = $u * $u2;
       $u4 = $u * $u3;
       $u5 = $u * $u4;
       $u6 = $u * $u5;

       $dTSec = 10583.6 - 1014.41*$u
                        + 33.78311*$u2
                        -  5.952053*$u3
                        -  0.1798452*$u4
                        +  0.022174192*$u5
                        +  0.0090316521*$u6;
      }

   if ($Y >= 500 and $Y < 1600)
      {
       $u  = ($y - 1000)/100;
       $u2 = $u * $u;
       $u3 = $u * $u2;
       $u4 = $u * $u3;
       $u5 = $u * $u4;
       $u6 = $u * $u5;

       $dTSec = 1574.2 - 556.01*$u
                       +  71.23472*$u2
                       +   0.319781*$u3
                       -   0.8503463*$u4
                       -   0.005050998*$u5
                       +   0.0083572073*$u6;
      }


   if ($Y >= 1600 and $Y < 1700)
      {
       $u  = $y - 1600;

      $dTSec = 120 - 0.9808*$u - 0.01532*$u*$u + $u*$u*$u/7129;
      }


   if ($Y >= 1700 and $Y < 1800)
      {
       $u  = $y - 1700;
       $u2 = $u * $u;
       $u3 = $u * $u2;
       $u4 = $u * $u3;

       $dTSec = 8.83 + 0.1603*$u
                     - 0.0059285*$u2
                     + 0.00013336*$u3
                     - $u4/1174000;
      }


   if ($Y >= 1800 and $Y < 1860)
      {
       $u  = $y - 1800;
       $u2 = $u * $u;
       $u3 = $u * $u2;
       $u4 = $u * $u3;
       $u5 = $u * $u4;
       $u6 = $u * $u5;
       $u7 = $u * $u6;

       $dTSec = 13.72 - 0.332447*$u
                      + 0.0068612*$u2
                      + 0.0041116*$u3
                      - 0.00037436*$u4
                      + 0.0000121272*$u5
                      - 0.0000001699*$u6
                      + 0.000000000875*$u7;
      }


   if ($Y >= 1860 and $Y < 1900)
      {
       $u  = $y - 1860;
       $u2 = $u * $u;
       $u3 = $u * $u2;
       $u4 = $u * $u3;
       $u5 = $u * $u4;

       $dTSec = 7.62 + 0.5737*$u
                     - 0.251754*$u2
                     + 0.01680668*$u3
                     - 0.0004473624*$u4
                     + $u5 / 233174;
      }


   if ($Y >= 1900 and $Y < 1920)
      {
       $u  = $y - 1900;
       $u2 = $u * $u;
       $u3 = $u * $u2;
       $u4 = $u * $u3;

       $dTSec = 1.494119*$u - 2.79
                            - 0.0598939*$u2
                            + 0.0061966*$u3
                            - 0.000197*$u4;
      }


   if ($Y >= 1920 and $Y < 1941)
      {
       $u  = $y - 1920;
       $u2 = $u * $u;
       $u3 = $u * $u2;

       $dTSec = 21.20 + 0.84493*$u - 0.076100*$u2 + 0.0020936*$u3;
      }


   if ($Y >= 1941 and $Y < 1961)
      {
       $u  = $y - 1950;
       $u2 = $u * $u;
       $u3 = $u * $u2;

       $dTSec = 29.07 + 0.407*$u - $u2/233 + $u3/2547;
      }


   if ($Y >= 1961 and $Y < 1986)
      {
       $u  = $y - 1975;
       $u2 = $u * $u;
       $u3 = $u * $u2;

       $dTSec = 45.45 + 1.067*$u - $u2/260 - $u3/718;
      }


   if ($Y >= 1986 and $Y < 2005)
      {
       $u  = $y - 2000;
       $u2 = $u * $u;
       $u3 = $u * $u2;
       $u4 = $u * $u3;
       $u5 = $u * $u4;

       $dTSec = 63.86 + 0.3345*$u
                      - 0.060374*$u2
                      + 0.0017275*$u3
                      + 0.000651814*$u4
                      + 0.00002373599*$u5;
      }


   if ($Y >= 2005 and $Y < 2050)
      {
       $u = $y - 2000;

       $dTSec = 62.92 + 0.32217*$u + 0.005589*$u*$u;
      }


   if ($Y >= 2050 and $Y <= 2150)
      {
       $u  = ($y - 1820) / 100;

       $dTSec = 32*$u*$u - 0.5628*(2150 - $y) - 20;
      }


   if ($Y > 2150 and $Y <= 3000)
      {
       $u  = ($y - 1820) / 100;

       $dTSec = 32*$u*$u - 20;
      }

// Format to nearest second.
   return SPrintF("%+d", round($dTSec, 0));

} // end OF  NASA_Delta_T_Poly (...)




/*
   ===========================================================================
   This function translates a purely numeric time string into standard form.
   Applies ONLY to purely numerical time strings.

   time format used = 00 to 24 hours

   ----------------------------------
               EXAMPLES
   time String          Translates to
   -----------          -------------
      1                    01:00:00
      12                   12:00:00
      123                  01:23:00
      1234                 12:34:00
      12345                01:23:45
      123456               12:34:56

   This function does NOT check for errors.

   NO DEPENDENCIES
   ===========================================================================
*/

   function Num_to_HMS ($NumHMSVal)
{
   $w = trim($NumHMSVal);

// Equate empty input to zero.
   if ($w == '') {return '00:00:00';}

// Return input unchanged if non-numeric argument
// rather than return an error message.
   if (!Is_Numeric($w)) {return $NumHMSVal;}

// Get first character and numerical sign.
   $FirstChar = substr($w,0,1);
   $NumSign   = (IntVal($w) < 0)? '-' : '+';

// Remove any numerical () sign
// and work with absolute value.
   $w = Str_Replace($NumSign, '', $w);

// Construct (w) into a 6-digit integer.
   if (StrLen($w) == 1) {$w = "0$w";}
   if (StrLen($w) == 2) {$w = $w."0000";}
   if (StrLen($w) == 3) {$w = "0$w";}
   if (StrLen($w) == 4) {$w = $w."00";}
   if (StrLen($w) == 5) {$w = "0$w";}

// Only keep (+) sign if it was in the original argument.
// Otherwise, a positive value will NOT be returned with
// a (+) sign. This does not effect the sign of negative
// values.
   if ($FirstChar <> '+' and $NumSign == '+') {$NumSign = '';}

// Done.  Return assumed time string equivalent ('hh:mm:ss') with colons.
   return $NumSign.substr($w,0,2).':'.substr($w,2,2).':'.substr($w,4,2);

} // End of  Num_to_HMS (...)




/*
   ============================================================================
   This function converts a time elements string ('h m s') to decimal hours.

   NO DEPENDENCIES
   ============================================================================
*/

   function HMS_to_Hours ($hhmmss)
{
// Read hour angle string argument and force to lower case in
// the event any (h m s :) symbols are used.
   $hms = StrToLower(trim($hhmmss));

// Convert any (h m s :) symbols into spaces.
   $hms = Str_Replace('h', ' ', $hms);
   $hms = Str_Replace('m', ' ', $hms);
   $hms = Str_Replace('s', ' ', $hms);
   $hms = Str_Replace(':', ' ', $hms);

// Normalize the spacing and then split and extract the
// individual time string elements (hh, mm,ss).
   $hms = PReg_Replace("/\s+/", " ", trim($hms));
   $whms = PReg_Split("[ ]", $hms);
   $whmscount = count($whms);
   $hh = ($whmscount >= 1)? bcAdd($whms[0],"0", 20) : "0";
   $mm = ($whmscount >= 2)? bcAdd($whms[1],"0", 20) : "0";
   $ss = ($whmscount >= 3)? bcAdd($whms[2],"0", 20) : "0";

// Remember and then remove any numerical (±) sign.
   $NumSign = (substr($hhmmss,0,1) == '-')? '-' : '';
   $hh = Str_Replace('-', '', $hh);
   $hh = Str_Replace('+', '', $hh);

// If original angle argument began with a + sign, then preserve
// it so that all returned positive results will have a + sign.
// Otherwise, positive results will NOT have a + sign.
   if (substr($hhmmss,0,1) == '+') {$NumSign = '+';}

// Compute decimal hours value equivalent
// to the combined hh,mm,ss values.
   $w2 = bcAdd(bcAdd(bcMul($hh,'3600',20),bcMul($mm,'60',20),20),$ss,20);

// If result equates to zero, then suppress any numerical sign.
   if (bcComp($w2, '0', 20) == 0) {$NumSign = '';}

// round off result to 16 decimals.
   $w =  bcAdd(bcDiv($w2,3600, 20), '0.00000000000000005', 16);

// Reattach numerical sign and trim off
// any redundant zeros/decimal point.
   return $NumSign.Rtrim(Rtrim($w, '0'), '.');

} // end of  HMS_to_Hours (hhmmss)






/*
   This function computes the BC/AD calendar date for any given
   Julian Date (JD) on the Julian or Gregorian calendar from
   scratch, not using any of the native PHP date functions.

   JDVal    = The Julian Date value
   CalMode  = Calendar mode: 'G' = Gregorian = Default,  'J' = Julian
   Decimals = Decimals used for seconds (0=Default  to  3)

   OUTPUT
   A date and time string such as
   AD 1949-May-20-Fri 09:02:13.334
   corresponding to the JD argument.

*/

   function Inverse_JD($JDVal, $CalMode='G', $Decimals=0)
{
  $JD    = trim($JDVal);
  $J     = $JD - 0.5;
  $CMode = (StrToUpper(substr($CalMode,0,1)) == 'J')? 0:1;

  $q = abs(IntVal(trim($Decimals)));   if ($q > 3) {$q = 3;}
  $q = '0'.(($q == 0)? 2 : $q+3).".$q".'f';

// Compute calendar date (BC/AD y, m, d) elements.
  $A = floor($JD + 0.5);
  $B = $CMode*floor(($A - 1867216.25) / 36524.25);
  $C = $A + $CMode*($B - floor($B/4) + 1);
  $D = $C + 1524;
  $E = floor(($D - 122.1) / 365.25);
  $F = floor(365.25 * $E);
  $G = floor(($D - $F) / 30.6001);

  $d = $D - $F - floor(30.6001 * $G);     // Day num   (1 to 31)
  $m = $G - 12*floor($G/14) - 1;          // Month num (1 to 12)
  $y = $E - 4716 + floor((14 - $m) / 12); // Math year num
  $y = ($y > 0)? $y : $y-1;               // Cal year num (Negative = BC)

// Construct BC/AD calendar date string.
  $Mmm = substr('JanFebMarAprMayJunJulAugSepOctNovDec', 3*($m-1), 3);
  $DoW = substr('SunMonTueWedThuFriSat', 3*(($A+1) % 7), 3);
  $ymd = (($y < 0)?'BC ':'AD ').abs($y)."-$Mmm-".SPrintF("%02d",$d)."-$DoW";

// Compute time of day (hh, mm, ss.sss)
// elements to the nearest millisecond.
  $hrs = 24*($J - floor($J));
  $hh  = floor($hrs);          // hour    hh
  $min = 60*($hrs - $hh);
  $mm  = floor($min);          // minutes mm
  $sec = 60*($min - $mm);
  $sss = SPrintF("%$q", $sec); // seconds ss.sss
  $ss  = floor($sec);          // seconds ss

// Patch for that blasted 60s glitch.
   if ($ss == 60 or $sss == 60)
      {$mm += 1;    $ss  = 0; $sss = SPrintF("%$q", 0);}
   if ($mm == 60)  {$hh += 1; $mm = 0;}

// Format the time (hh, mm, ss) elements.
  $hh = SPrintF("%02d", $hh);
  $mm = SPrintF("%02d", $mm);
  $ss = SPrintF("%02d", $ss);

// Construct and return final date and time string.
   return "$ymd $hh:$mm:$sss";
} // End of  Inverse_JD()





?>


